## Games - Video/Audio modes and extra functionality
In the wild west of x68000 game development, there doesn't seem to be much in the way of standards across games. This can include switching to 31Khz or turning on MIDI audio or different controls. Luckily someone [before me](https://vintagecpu.wordpress.com/2010/12/31/switching-resolution-for-x68000-vga/) has gone through and documented much of this very important information, so important that it would be a shame for it to disappear one day. This page serves the purpose of a backup, and an extension of this information.

## Settings and Notes

Name                     | Default Video Mode  | Optional Video Mode(s)  | Notes
---|:---:|:---:|:--------------------------------------------------
| After Burner             | 15kHz  | 31kHz  | Mouse, Trackball, Micom Analogue pad, Cyber Stick (CZ-8NJ2) support. Mousepad included. On machines with 2MB of RAM or more, hold SHIFT while loading to load the entire game into memory.  |
| Ajax                     | 31kHz  | 15kHz - hold down the HELP key when first booting | Ajax is a three-button game, and can be played as such using the keyboard. Using a joystick reduces the number of buttons to two.  |
| Akumajou Dracula         | 31kHz  |        | The famous X68000 version of Castlevania. Requires a blank disk for saving, or to hard disk. At the music selection screen, press the XF1 and G keys together to enable GM music.  |
| Alien Syndrome           | 31kHz  | 15kHz  |   |
| Argus no Senshi (Rygar)  | 31kHz  |        |   |
| Arkanoid Revenge of DOH  | 31kHz  | 15kHz – hold down the OPT1 key when first booting  | 16MHz runs too fast. Mouse/Trackball support.  |
| Atomic Robo-Kid          | 15kHz  |        | For MIDI support, hold down F1 when booting  |
| Baraduke                 | 31kHz  | 15kHz, 24kHz – press HELP to switch resolutions  | On the high score table, enter your name as JUNKO, KISSY or TAKKY and observe the result.  |
| Battletech               | 31kHz  |        |   |
| Bomberman                | 31kHz  |        |   |
| Bonanza Brothers         | 31kHz  | 15kHz  |   |
| Bosconian                | 31kHz  |        | To advance to the next stage, press the Q, E, T and U keys together while playing the game. If you turn the code key on, you'll hear an echo of the sampled voice.  |
| Bubble Bobble            | 31kHz  | 15kHz  |   |
| Butasan                  | 31kHz  | 15kHz – press HELP to switch resolutions  |   |
| Chelnov                  | 31kHz  | 15kHz – press HELP to switch resolutions  | Megadrive joypad adapter included, as three buttons are required  |
| Chourensha               | 31kHz (Narrow)  | 31kHz (Wide), 15kHz, The display can be rotated  | Free downloadable doujinsoft  |
| Chuka Taisen             | 15kHz  | 31kHz - press F1 to alternate  |   |
| Code Zero                | 15kHz  |        |   |
| Columns                  | 31kHz  |        |   |
| Cotton                   | 31kHz (384 x 256)  | 31kHz (512 x 256), 15kHz (448 x 240), 15kHz (512 x 240)  |   |
| Crazy Climber 1/2        | 31kHz  |        | Crazy Climber 2 requires 2 MB, makes use of the XPD-1LR joypad (included with Libble Rabble). To load Crazy Climber 2, boot the game from Drive 1.  |
| Cybercore                | 31kHz  |        | Can be run from hard drive, as long as the disk is in Drive 0. The program name is “Y.x”  |
| D-Return                 | 31kHz  |        | 16MHz runs too fast  |
| DaiMakaiMura (Ghouls’n Ghosts)  | 15kHz  | 31kHz  | Menu navigation to switch video modes:  |
|                          ||| 1. Button 1  |
|                          ||| 2. Down  |
|                          ||| 3. Button 1  |
|                          ||| 4. Down 4 times  |
|                          ||| 5. Left/Right to select resolution  |
| Dash Yarou               | 31kHz  |        | Cyber Stick (CZ-8NJ2) support  |
| Detana!! Twinbee         | 31kHz  | 15kHz – use F2 to select, F1 to revert to 31kHz  | Pressing [ESC] in-game and then F3 enables a pseudo-vertical mode. If you have more than 2MB of memory, hold down HELP while booting to load the entire game into memory. This has a music mode like Parodius. Run SD_DRV.BAT from the command line.  |
| Dig Dug/Dig Dug II       | 31kHz  |        | To load Dig Dug II, boot the game from Drive 1  |
| Dive On                  | 31kHz (384 x 256)  | 31kHz (256 x 256), 15kHz (256 x 256), 15kHz (512 x 256)  |   |
| Downtown Nekketsu Monogatari (River City Ransom)  | 31kHz  | 15kHz – hold down the HELP key when first booting  |   |
| Dragon Buster            | 31kHz  | 24kHz – press HELP to switch resolutions  |   |
| Dragon Spirit            | 31kHz (256 x 256)  | 31kHz (512 x 256), 15kHz (256 x 256), 15kHz (512 x 256) The display can be rotated  | Run in 10MHz or music is messed up. Some releases includes a flag. Try setting the internal clock to between midnight and 1:00AM  |
| Dungeon Master           | 31kHz  |        |   |
| Dynamite Duke            | 31kHz  |        |   |
| Etoile Princesse         | 31kHz  |        | Not MT-32 compatible. Requires a blank disk for saving.  |
| Exciting Hour/Shusse Oozumou  | 31kHz  | 15kHz – press HELP to switch resolutions  | To load Shusse Oozumou, boot the game from Drive 1  |
| F-15 Strike Eagle II     | 31kHz  |        | 16MHz highly recommended  |
| Fairyland Story          | 15kHz  | 31kHz  | To see the staff credits, turn Continue off on the options screen, start a game and when the Game Over message appears, press the A and B buttons on your joystick at the same time.  |
| Fantasy Zone             | 31kHz  | 24kHz – press HELP to switch resolutions  | Run in 10MHz or music is messed up. Fantasy Zone includes the undocumented Fantasy Zone 3D - hold down the Shift, CTRL and HELP keys together when booting to activate it. To fast forward through the super play, hold down the CTRL key.  |
| Final Fight              | 15kHz (512 x 256)  | 31kHz (512 x 256), 31kHz (384 x 256)  | Menu Navigation to switch video modes:  |
|                          ||| 1. Button 1  |
|                          ||| 2. Down  |
|                          ||| 3. Button 1  |
|                          ||| 4. Down 5 times  |
|                          ||| 5. Left/Right to select resolution  |
|                          ||| Special soundtrack CD included  |
| Flying Shark             | 31kHz  | 15kHz  |   |
| Full Throttle            | 31kHz  |        | 16MHz runs too fast  |
| FZ Senki Axis            | 31kHz  | 15kHz  |   |
| Galaga ’88               | 31kHz  | 15kHz  |   |
| Garou Densetsu (Fatal Fury)  | 31kHz  |   | Try setting the internal clock to between midnight and 1:00AM  |
| Garou Densetsu 2 (Fatal Fury 2)  | 31kHz  | 15kHz  | Joypad included  |
| Garou Densetsu Special (Fatal Fury Special)  | 31kHz  | 15kHz  |   |
| Gemini Wing              | 31kHz  | 15kHz  | For MIDI support hold down while booting:  |
|                          ||| 1. F1: MT32 or CM32L  |
|                          ||| 2. F2: CM64  |
|                          ||| 3. F3: CM64 + SN-U110-10 expansion  |
|                          ||| 4. F4: CM64 + SN-U110-07 expansion  |
|                          ||| Infinite continues: on the title screen, hold down the G, E, M, I, N keys at the same time and you'll be told you have 9 continues. In reality, they should be infinite. Gemini Wing also has a vertical mode, hold down F7 while the game is booting. Unfortunately it doesn't allow you to rotate the monitor.  |
| Gempai Toumaden          | 31kHz  |        | Some releases includes a flag and flagpole  |
| Genocide                 | 31kHz  |        |   |
| Genocide 2               | 31kHz  | 15kHz  |   |
| Geograph Seal            | 31kHz  |   | Not MT-32 compatible. Requires a blank disk for saving.  |
| Gradius                  | 31kHz  |   | Run in 10MHz or music is messed up. Can be run from a hard disk by simply copying all files to it and running Gradius.x  |
| Gradius II               | 31kHz  |   | Although Gradius II has its real arcade mode hidden but documented, what's not documented is the arcade test mode. Enable it by holding down F4 while booting (at the same time as holding down the button to enable arcade mode).  |
| Granada                  | 31kHz  | 15kHz  | 16MHz runs too fast. For MIDI support hold down the 登録 key while booting.  |
| Gunship                  | 31kHz  |        |   |
| Image Fight              | 31kHz  | 15kHz  |   |
| Jack Nicklaus Golf       | 31kHz  |        |   |
| Klax                     | 31kHz  |        |   |
| KU2                      | 15kHz  | 31kHz  |   |
| Kyuukyoku Tiger (Twin Cobra)  | 31kHz  |   | Kyuukyoku Tiger can be controlled with the mouse. Hold down F2 while booting to enable it.  |
| Lemmings                 | 31kHz  |   |   |
| Libble Rabble            | 31kHz  | 15kHz, 24kHz – press HELP to switch resolutions  | Includes special joypad (XPD-1LR), can also be used by Crazy Climber 1/2  |
| Mahou Daisakusen (Sorcer Striker)  | 31kHz (256 x 256)  | 31kHz (384 x 256), 15kHz (256 x 240)  | 16MHz highly recommended  |
| Maison Ikkoku Kanketsuhen Special  | 31kHz  |   |   |
| Marble Madness           | 31kHz (384 x 512)  | 24kHz (336 x 448)  | Trackball support. A real marble was supplied with the game.  |
| Märchen Maze             | 15kHz (256 x 224)  | 31kHz (288 x 224), 31kHz (256 x 224), 15kHz (256 x 224) Press the 登録 key to alternate  | On the title screen of Marchen Maze, press Shift and HELP together to access the options screen, allowing you among other things to select the Arcade version. Bonus trick: Start the game with Disk B in Drive 0 and Disk A in drive 1 to activate PCM sound.  |
| Metal Sight              | 15kHz  | 31kHz  | Cyber Stick (CZ-8NJ2) support  |
| Mid-Garts                | 31kHz  |   |   |
| Motos                    | 31kHz (256 x 256)  | 31kHz (384 x 256)  | On Christmas Eve and New Year's Day different opening music is played. For an arranged soundtrack, instead of loading the game from drive 0, load it from drive 1. If you turn on the 「ひらがな」「全角」「INS」「CAPS」「コード入力」「ローマ字」「かな」 keys within the config screen, you can select a special SARU mode, which is easier than the EASY mode.  |
| Mr. Do! / Mr. Do! V.S. Unicorns  | 31kHz  | 15kHz – press HELP to switch resolutions  | To load Mr. Do! V.S Unicorns, boot the game from Drive 1.  |
| Naious                   | 31kHz  | 15kHz – press HELP to switch resolutions  | Run in 10MHz or music is messed up. On the title screen, hold down [CTRL] and the W, I, N, K keys and you can select any stage from 1 to 7.  |
| Nekketsu Koukou Dodgeball Bu  | 31kHz  | 15kHz – hold down the HELP key when first booting  | Run in 10MHz or corrupt graphics.  |
| Soccer Hen               | 31kHz  |            |   |
| Nemesis ’90 Kai  | 31kHz  |       |       |
| Overtake                 | 31kHz (256)  | 31kHz (256 SQ)15kHz (256 Low)  | Cyber Stick (CZ-8NJ2) support  |
| Pac-Land                 | 31kHz  |   | Includes a three-button pad  |
| Pac-Mania                | 31kHz  | 15kHz – hold down the HELP key when first booting  | From the Pac-Mania title screen, press F5 to redefine the keys. Redefine them to X,6,8,0,0,0 and make sure to use the numeric keypad. Exit by pressing Esc and you'll automatically enter the hidden options screen. Bonus game: The original Pac-Man is included, run the pac.bat batch file in the PACMAN folder.  |
| Parodius                 | 31kHz  |       | There's a sound driver test mode on Disk A. Run SD_DRV.X from the command line. As the program tells you, pressing Space progresses to the next piece and ESC will end the program. For fun, try holding down the following keys while the game boots:  |
|                          ||| • K, 6, 8 : one for the MSX fans probably  |
|                          ||| • [SPACE]  |
|                          ||| • [HOME], G, 2  |
| Phalanx                  | 31kHz  | 15kHz  | Phalanx includes a hidden stage. Crash into what looks like a red window and you'll be warped through. On the configuration screen, the following commands can be entered for their respective effects (For these to work, you have to hold down the first key, and then the second together, then take your finger off the first key and press the third and so on. So for the first cheat you'd hold A and K, then K and I, then I and C and so on):  |
|                          ||| • AKICHANSUKI - Stage Select  |
|                          ||| • SUGOINO - Invincibility  |
|                          ||| • FUERUNO - Increase your number of lives  |
| Populous                 | 31kHz  |        |   |
| Populous The Promised Lands  | 31kHz  |   | Add-on disk for Populous  |
| Populous 2               | 31kHz  |        |   |
| Powermonger              | 31kHz  |        |   |
| Prince of Persia         | 31kHz  |        |   |
| Pro Soccer 68            | 15kHz  | 31kHz – press HELP to switch resolutions  |   |
| Puyo Puyo                | 15kHz  | 31kHz (Six resolution options to choose from)  | To enable Tokoton Puyo Puyo mode, at the title screen push the second player's controller to the bottom left and press F1.  |
| Puzznic                  | 31kHz  |        |   |
| Quarth                   | 31kHz  |        |   |
| Quiz Torimonochou        | 31kHz  | 15kHz, 24kHz  |   |
| R-Type                   | 31kHz  |        | You can start on any level in R-Type. Select the level you want to play from the DEMO option, and then start the game while holding down the S, T, A, G, and E keys.  |
| R.C. Robot Construction  | 31kHz  |        |   |
| Salamander               | 31kHz  | 15kHz – hold down the HELP key when first booting  |   |
| Scorpius                 | 31kHz  |        |   |
| Shoot Range              | 31kHz  |        |   |
| Sim City                 | 31kHz  |        |   |
| Sim Earth                | 31kHz  |        | Requires SX-Window to run  |
| Sol-Feace                | 31kHz  | 15kHz  | For MIDI hold 登録 while booting  |
| Space Harrier            | 15kHz  |        |   |
| Star Force               | 31kHz  | 15kHz – press HELP to switch resolutions  |   |
| Star Luster              | 15kHz  | 31kHz  | Not MT-32 compatible. Supports the AE-1AP Analogue Joypad  |
| Star Trader              | 31kHz  | 15kHz  | A Takeru game - not a full commercial release  |
| Star Wars Attack on the Death Star  | 31kHz  |   | Runs better at 16 MHz  |
| Street Fighter II’       | 15kHz  | 31kHz  | Menu navigation to switch video modes:  |
|                          ||| 1. Button 1  |
|                          ||| 2. Down  |
|                          ||| 3. Button A  |
|                          ||| 4. Down  |
|                          ||| 5. Right  |
|                          ||| Includes an adapter for Megadrive controllers, specifically the Capcom Power Stick Fighter. Remap for Genesis buttons:  |
|                          ||| • KICK0: C  |
|                          ||| • KICK1: A  |
|                          ||| • KICK2: B  |
| Strider Hiryuu           | 15kHz (512 x 256)  | 31kHz (512 x 256), 31kHz (384 x 256)  | Menu navigation to switch video modes:  |
|                          ||| 1. Button 1  |
|                          ||| 2. Down  |
|                          ||| 3. Button 1  |
|                          ||| 4. Down 4 times  |
|                          ||| 5. Left/Right to select resolution  |
| Super Hang-On            | 15kHz  | 31kHz – hold down the HELP key when first booting  | Cyber Stick (CZ-8NJ2) support  |
| Super Street Fighter II  | 15kHz (Narrow)  | 15kHz (Wide), 31kHz (Narrow), 31kHz (Wide)  | 16 MHz recommended. Remap for Genesis buttons:  |
|                          ||| 1. KICK0: C  |
|                          ||| 2. KICK1: A  |
|                          ||| 3. KICK2: B  |
| Syvalion                 | 15kHz  | 31kHz – hold down the HELP key when first booting  | Trackball support. Cyber Stick (CZ-8NJ2) support  |
| Terra Cresta & Moon Cresta  | 31kHz  | You can switch between narrow and wide display sizes by pressing the HELP key  | To load Moon Cresta, boot the game from Drive 1  |
| Tetris                   | 31kHz  |       |   |
| The NewZealand Story     | 31kHz  | 15kHz – hold down the HELP key when first booting  |   |
| The Return of Ishtar     | 15kHz  |       |   |
| Thunder Blade            | 31kHz  |       | Cyber Stick (CZ-8NJ2) support  |
| Thunderforce II          | 31kHz  | 15kHz – hold down the F10 key when first booting  |   |
| Twinbee                  | 15kHz  |       | Run in 10MHz or corrupt graphics  |
| Undeadline               | 31kHz  |       |   |
| V’Ball                   | 31kHz  | 15kHz – hold down the HELP key when first booting  |   |
| Valis II                 | 31kHz  |       | Run in 10MHz or music is messed up.  |
| Viewpoint                | 31kHz  |       | Runs better at 16 MHz, but it’s still slow compared to the Neo Geo version  |
| World Court Pro Tennis   | 15kHz  | 31kHz (Square), 31kHz (256 x 256), 15kHz (256 x 256)  | Hold Down 0, 1 or 2 on the numeric keypad respectively while booting  |
| Xenon 2                  | 31kHz  | 15kHz, 24kHz – press HELP to switch resolutions  |   |
| Xevious                  | 31kHz  |       | Both Xevious and Super Xevious are available  |


###### Original document source is from [NFG / GameSX](https://gamesx.com/wiki/doku.php?id=x68000:game_notes), duplicated here to serve as backup and reference.
